from __future__ import absolute_import

import base64
import json
import logging
import time

import six
import msgpack

from ably.types.typedbuffer import TypedBuffer
from ably.types.mixins import EncodeDataMixin
from ably.util.crypto import CipherData
from ably.util.exceptions import AblyException

log = logging.getLogger(__name__)


class Message(EncodeDataMixin):
    def __init__(self, name=None, data=None, client_id=None,
                 id=None, connection_id=None, timestamp=None,
                 encoding=''):
        if name is None:
            self.__name = None
        elif isinstance(name, six.string_types):
            self.__name = name
        elif isinstance(name, six.binary_type):
            self.__name = name.decode('ascii')
        else:
            # log.debug(name)
            # log.debug(name.__class__)
            raise ValueError("name must be a string or bytes")
        self.__id = id
        self.__client_id = client_id
        self.__data = data
        self.__timestamp = timestamp
        self.__connection_id = connection_id
        super(Message, self).__init__(encoding)

    def __eq__(self, other):
        if isinstance(other, Message):
            return (self.name == other.name
                    and self.data == other.data
                    and self.client_id == other.client_id
                    and self.timestamp == other.timestamp)
        return NotImplemented

    def __ne__(self, other):
        if isinstance(other, Message):
            result = self.__eq__(other)
            if result != NotImplemented:
                return not result
        return NotImplemented

    @property
    def name(self):
        return self.__name

    @property
    def client_id(self):
        return self.__client_id

    @property
    def data(self):
        return self.__data

    @property
    def connection_id(self):
        return self.__connection_id

    @property
    def id(self):
        return self.__id

    @property
    def timestamp(self):
        return self.__timestamp

    def encrypt(self, channel_cipher):
        if isinstance(self.data, CipherData):
            return

        elif isinstance(self.data, six.text_type):
            self._encoding_array.append('utf-8')

        if isinstance(self.data, dict) or isinstance(self.data, list):
            self._encoding_array.append('json')
            self._encoding_array.append('utf-8')

        typed_data = TypedBuffer.from_obj(self.data)
        if typed_data.buffer is None:
            return True
        encrypted_data = channel_cipher.encrypt(typed_data.buffer)
        self.__data = CipherData(encrypted_data, typed_data.type,
                                 cipher_type=channel_cipher.cipher_type)

    @staticmethod
    def decrypt_data(channel_cipher, data):
        if not isinstance(data, CipherData):
            return
        decrypted_data = channel_cipher.decrypt(data.buffer)
        decrypted_typed_buffer = TypedBuffer(decrypted_data, data.type)

        return decrypted_typed_buffer.decode()

    def decrypt(self, channel_cipher):
        decrypted_data = self.decrypt_data(channel_cipher, self.__data)
        if decrypted_data is not None:
            self.__data = decrypted_data

    def as_dict(self, binary=False):
        data = self.data
        data_type = None
        encoding = self._encoding_array[:]

        if isinstance(data, dict) or isinstance(data, list):
            encoding.append('json')
            data = json.dumps(data)
        elif isinstance(data, six.text_type) and not binary:
            # text_type is always a unicode string
            pass
        elif (not binary and isinstance(data, bytearray) or
              # bytearray is always bytes
              isinstance(data, six.binary_type) and six.binary_type != str):
                # in py3k we will understand <class 'bytes'> as bytes
            data = base64.b64encode(data).decode('ascii')
            encoding.append('base64')
        elif isinstance(data, CipherData):
            encoding.append(data.encoding_str)
            data_type = data.type
            if not binary:
                data = base64.b64encode(data.buffer).decode('ascii')
                encoding.append('base64')
            else:
                data = data.buffer
        elif binary and isinstance(data, bytearray):
            data = six.binary_type(data)

        if not (isinstance(data, (six.binary_type, six.text_type, list, dict,
                                  bytearray)) or
                data is None):
            raise AblyException("Invalid data payload", 400, 40011)

        request_body = {
            'name': self.name,
            'data': data,
            'timestamp': self.timestamp or int(time.time() * 1000.0),
        }
        request_body = {k: v for (k, v) in request_body.items()
                        if v is not None}  # None values aren't included

        if encoding:
            request_body['encoding'] = '/'.join(encoding).strip('/')

        if data_type:
            request_body['type'] = data_type

        if self.client_id:
            request_body['clientId'] = self.client_id

        if self.id:
            request_body['id'] = self.id

        if self.connection_id:
            request_body['connectionId'] = self.connection_id

        return request_body

    def as_json(self):
        return json.dumps(self.as_dict(), separators=(',', ':'))

    @staticmethod
    def from_dict(obj, cipher=None):
        id = obj.get('id')
        name = obj.get('name')
        data = obj.get('data')
        client_id = obj.get('clientId')
        connection_id = obj.get('connectionId')
        timestamp = obj.get('timestamp')
        encoding = obj.get('encoding', '')

        decoded_data = Message.decode(data, encoding, cipher)

        return Message(
            id=id,
            name=name,
            connection_id=connection_id,
            client_id=client_id,
            timestamp=timestamp,
            **decoded_data
        )

    def as_msgpack(self):
        return msgpack.packb(self.as_dict(binary=True), use_bin_type=True)


def make_message_response_handler(binary):
    def message_response_handler(response):
        messages = response.to_native()
        return [Message.from_dict(j) for j in messages]
    return message_response_handler


def make_encrypted_message_response_handler(cipher, binary):
    def encrypted_message_response_handler(response):
        messages = response.to_native()
        return [Message.from_dict(j, cipher) for j in messages]
    return encrypted_message_response_handler


class MessageJSONEncoder(json.JSONEncoder):
    def default(self, message):
        if isinstance(message, Message):
            return message.as_dict()
        else:
            return json.JSONEncoder.default(self, message)



from __future__ import absolute_import

from ably.types.authoptions import AuthOptions
from ably.util.exceptions import AblyException


class Options(AuthOptions):
    def __init__(self, client_id=None, log_level=0, tls=True, rest_host=None,
                 realtime_host=None, port=0, tls_port=0, use_binary_protocol=True,
                 queue_messages=False, recover=False, environment=None,
                 http_open_timeout=None, http_request_timeout=None,
                 http_max_retry_count=None, http_max_retry_duration=None,
                 **kwargs):
        super(Options, self).__init__(**kwargs)

        # TODO check these defaults

        self.__client_id = client_id
        self.__log_level = log_level
        self.__tls = tls
        self.__rest_host = rest_host
        self.__realtime_host = realtime_host
        self.__port = port
        self.__tls_port = tls_port
        self.__use_binary_protocol = use_binary_protocol
        self.__queue_messages = queue_messages
        self.__recover = recover
        self.__environment = environment
        self.__http_open_timeout = http_open_timeout
        self.__http_request_timeout = http_request_timeout
        self.__http_max_retry_count = http_max_retry_count
        self.__http_max_retry_duration = http_max_retry_duration

    @property
    def client_id(self):
        return self.__client_id

    @client_id.setter
    def client_id(self, value):
        self.__client_id = value

    @property
    def log_level(self):
        return self.__log_level

    @log_level.setter
    def log_level(self, value):
        self.__log_level = value

    @property
    def tls(self):
        return self.__tls

    @tls.setter
    def tls(self, value):
        self.__tls = value

    @property
    def rest_host(self):
        return self.__rest_host

    @rest_host.setter
    def rest_host(self, value):
        self.__rest_host = value

    @property
    def realtime_host(self):
        return self.__realtime_host

    @realtime_host.setter
    def realtime_host(self, value):
        self.__realtime_host = value

    @property
    def port(self):
        return self.__port

    @port.setter
    def port(self, value):
        self.__port = value

    @property
    def tls_port(self):
        return self.__tls_port

    @tls_port.setter
    def tls_port(self, value):
        self.__tls_port = value

    @property
    def use_binary_protocol(self):
        return self.__use_binary_protocol

    @use_binary_protocol.setter
    def use_binary_protocol(self, value):
        self.__use_binary_protocol = value

    @property
    def queue_messages(self):
        return self.__queue_messages

    @queue_messages.setter
    def queue_messages(self, value):
        self.__queue_messages = value

    @property
    def recover(self):
        return self.__recover

    @recover.setter
    def recover(self, value):
        self.__recover = value

    @property
    def environment(self):
        return self.__environment

    @property
    def http_open_timeout(self):
            return self.__http_open_timeout

    @http_open_timeout.setter
    def http_open_timeout(self, value):
        self.__http_open_timeout = value

    @property
    def http_request_timeout(self):
            return self.__http_request_timeout

    @http_request_timeout.setter
    def http_request_timeout(self, value):
        self.__http_request_timeout = value

    @property
    def http_max_retry_count(self):
            return self.__http_max_retry_count

    @http_max_retry_count.setter
    def http_max_retry_count(self, value):
        self.__http_max_retry_count = value

    @property
    def http_max_retry_duration(self):
            return self.__http_max_retry_duration

    @http_max_retry_duration.setter
    def http_max_retry_duration(self, value):

        self.__http_max_retry_duration = value



from __future__ import absolute_import

import base64
import logging
import time
import uuid

import six
import requests

from ably.types.capability import Capability
from ably.types.tokendetails import TokenDetails
from ably.types.tokenrequest import TokenRequest
from ably.util.exceptions import AblyException, IncompatibleClientIdException

__all__ = ["Auth"]

log = logging.getLogger(__name__)


class Auth(object):

    class Method:
        BASIC = "BASIC"
        TOKEN = "TOKEN"

    def __init__(self, ably, options):
        self.__ably = ably
        self.__auth_options = options
        self.__client_id = options.client_id
        self.__client_id_validated = False

        self.__basic_credentials = None
        self.__auth_params = None
        self.__token_details = None

        must_use_token_auth = options.use_token_auth is True
        must_not_use_token_auth = options.use_token_auth is False
        can_use_basic_auth = options.key_secret is not None and options.client_id is None
        if not must_use_token_auth and can_use_basic_auth:
            # We have the key, no need to authenticate the client
            # default to using basic auth
            log.debug("anonymous, using basic auth")
            self.__auth_mechanism = Auth.Method.BASIC
            basic_key = "%s:%s" % (options.key_name, options.key_secret)
            basic_key = base64.b64encode(basic_key.encode('utf-8'))
            self.__basic_credentials = basic_key.decode('ascii')
            return
        elif must_not_use_token_auth and not can_use_basic_auth:
            raise ValueError('If use_token_auth is False you must provide a key')

        # Using token auth
        self.__auth_mechanism = Auth.Method.TOKEN

        if options.token_details:
            self.__token_details = options.token_details
        elif options.auth_token:
            self.__token_details = TokenDetails(token=options.auth_token)
        else:
            self.__token_details = None

        if options.auth_callback:
            log.debug("using token auth with auth_callback")
        elif options.auth_url:
            log.debug("using token auth with auth_url")
        elif options.key_secret:
            log.debug("using token auth with client-side signing")
        elif options.auth_token:
            log.debug("using token auth with supplied token only")
        elif options.token_details:
            log.debug("using token auth with supplied token_details")
        else:
            raise ValueError("Can't authenticate via token, must provide "
                             "auth_callback, auth_url, key, token or a TokenDetail")

    def authorise(self, token_params=None, auth_options=None, force=False):
        self.__auth_mechanism = Auth.Method.TOKEN

        if token_params is None:
            token_params = dict(self.auth_options.default_token_params)
        else:
            token_params = dict(self.auth_options.default_token_params,
                                **token_params)
            self.auth_options.default_token_params = dict(token_params)
            self.auth_options.default_token_params.pop('timestamp', None)

        if auth_options is not None:
            force = auth_options.pop('force', None) or force
            self.auth_options.merge(auth_options)
        auth_options = dict(self.auth_options.auth_options)
        token_params.setdefault('client_id', self.client_id)

        if self.__token_details:
            if not self.__token_details.is_expired(self._timestamp()):
                if not force:
                    log.debug(
                        "using cached token; expires = %d",
                        self.__token_details.expires
                    )
                    return self.__token_details
            else:
                # token has expired
                self.__token_details = None

        self.__token_details = self.request_token(token_params, **auth_options)
        self._configure_client_id(self.__token_details.client_id)
        return self.__token_details

    def request_token(self, token_params=None,
                      # auth_options
                      key_name=None, key_secret=None, auth_callback=None,
                      auth_url=None, auth_method=None, auth_headers=None,
                      auth_params=None, query_time=None):
        token_params = token_params or {}
        token_params = dict(self.auth_options.default_token_params,
                            **token_params)
        key_name = key_name or self.auth_options.key_name
        key_secret = key_secret or self.auth_options.key_secret

        log.debug("Auth callback: %s" % auth_callback)
        log.debug("Auth options: %s" % six.text_type(self.auth_options))
        if query_time is None:
            query_time = self.auth_options.query_time
        query_time = bool(query_time)
        auth_callback = auth_callback or self.auth_options.auth_callback
        auth_url = auth_url or self.auth_options.auth_url

        auth_params = auth_params or self.auth_options.auth_params or {}

        auth_method = (auth_method or self.auth_options.auth_method).upper()

        auth_headers = auth_headers or self.auth_options.auth_headers or {}

        log.debug("Token Params: %s" % token_params)
        if auth_callback:
            log.debug("using token auth with authCallback")
            token_request = auth_callback(token_params)
        elif auth_url:
            log.debug("using token auth with authUrl")

            token_request = self.token_request_from_auth_url(
                auth_method, auth_url, token_params, auth_headers, auth_params)
        else:
            token_request = self.create_token_request(
                token_params, key_name=key_name, key_secret=key_secret,
                query_time=query_time)
        if isinstance(token_request, TokenDetails):
            return token_request
        elif isinstance(token_request, dict) and 'issued' in token_request:
            return TokenDetails.from_dict(token_request)
        elif isinstance(token_request, dict):
            token_request = TokenRequest(**token_request)
        elif isinstance(token_request, six.text_type):
            return TokenDetails(token=token_request)
        # python2
        elif isinstance(token_request, six.binary_type) and six.binary_type == str:
            return TokenDetails(token=token_request)

        token_path = "/keys/%s/requestToken" % token_request.key_name

        response = self.ably.http.post(
            token_path,
            headers=auth_headers,
            native_data=token_request.to_dict(),
            skip_auth=True
        )

        AblyException.raise_for_response(response)
        response_dict = response.to_native()
        log.debug("Token: %s" % str(response_dict.get("token")))
        return TokenDetails.from_dict(response_dict)

    def create_token_request(self, token_params=None,
                             key_name=None, key_secret=None, query_time=None):
        token_params = token_params or {}
        token_request = {}

        key_name = key_name or self.auth_options.key_name
        key_secret = key_secret or self.auth_options.key_secret
        if not key_name or not key_secret:
            log.debug('key_name or key_secret blank')
            raise AblyException("No key specified: no means to generate a token", 401, 40101)

        token_request['key_name'] = key_name
        if token_params.get('timestamp'):
            token_request['timestamp'] = token_params['timestamp']
        else:
            if query_time is None:
                query_time = self.auth_options.query_time
            if query_time:
                token_request['timestamp'] = self.ably.time()
            else:
                token_request['timestamp'] = self._timestamp()

        token_request['timestamp'] = int(token_request['timestamp'])

        token_request['ttl'] = token_params.get('ttl') or TokenDetails.DEFAULTS['ttl']

        if token_params.get('capability') is None:
            token_request["capability"] = ""
        else:
            token_request['capability'] = six.text_type(
                Capability(token_params['capability'])
            )

        token_request["client_id"] = (
            token_params.get('client_id') or self.client_id)

        # Note: There is no expectation that the client
        # specifies the nonce; this is done by the library
        # However, this can be overridden by the client
        # simply for testing purposes
        token_request["nonce"] = token_params.get('nonce') or self._random_nonce()

        token_request = TokenRequest(**token_request)

        if token_params.get('mac') is None:
            # Note: There is no expectation that the client
            # specifies the mac; this is done by the library
            # However, this can be overridden by the client
            # simply for testing purposes.
            token_request.sign_request(key_secret.encode('utf8'))
        else:
            token_request.mac = token_params['mac']

        return token_request

    @property
    def ably(self):
        return self.__ably

    @property
    def auth_mechanism(self):
        return self.__auth_mechanism

    @property
    def auth_options(self):
        return self.__auth_options

    @property
    def auth_params(self):
        return self.__auth_params

    @property
    def basic_credentials(self):
        return self.__basic_credentials

    @property
    def token_credentials(self):
        if self.__token_details:
            token = self.__token_details.token
            token_key = base64.b64encode(token.encode('utf-8'))
            return token_key.decode('ascii')

    @property
    def token_details(self):
        return self.__token_details

    @property
    def client_id(self):
        return self.__client_id

    def _configure_client_id(self, new_client_id):
        # If new client ID from Ably is a wildcard, but preconfigured clientId is set,
        # then keep the existing clientId
        if self.client_id != '*' and new_client_id == '*':
            self.__client_id_validated = True
            return

        # If client_id is defined and not a wildcard, prevent it changing, this is not supported
        if self.client_id is not None and self.client_id != '*' and new_client_id != self.client_id:
            raise IncompatibleClientIdException(
                "Client ID is immutable once configured for a client. "
                "Client ID cannot be changed to '{}'".format(new_client_id), 400, 40012)

        self.__client_id_validated = True
        self.__client_id = new_client_id

    def can_assume_client_id(self, assumed_client_id):
        if self.__client_id_validated:
            return self.client_id == '*' or self.client_id == assumed_client_id
        elif self.client_id is None or self.client_id == '*':
            return True  # client ID is unknown
        else:
            return self.client_id == assumed_client_id

    def _get_auth_headers(self):
        if self.__auth_mechanism == Auth.Method.BASIC:
            return {
                'Authorization': 'Basic %s' % self.basic_credentials,
            }
        else:
            self.authorise()
            return {
                'Authorization': 'Bearer %s' % self.token_credentials,
            }

    def _timestamp(self):
        """Returns the local time in milliseconds since the unix epoch"""
        return int(time.time() * 1000)

    def _random_nonce(self):
        return uuid.uuid4().hex[:16]

    def token_request_from_auth_url(self, method, url, token_params,
                                    headers, auth_params):
        if method == 'GET':
            body = {}
            params = dict(auth_params, **token_params)
        elif method == 'POST':
            params = {}
            body = dict(auth_params, **token_params)

        from ably.http.http import Response
        response = Response(requests.request(
            method, url, headers=headers, params=params, data=body))

        AblyException.raise_for_response(response)
        try:
            token_request = response.to_native()
        except ValueError:
            token_request = response.text
        return token_request



import six


class UnicodeMixin(object):
    if six.PY3:
        def __str__(self):
            return self.__unicode__()
    else:
        def __str__(self):
            return self.__unicode__().encode('utf8')



import logging

try:
    from logging import NullHandler
except ImportError:
    class NullHandler(logging.Handler):
        def emit(self, record):
            pass

        def handle(self, record):
            pass

        def createLock(self):
            return None

logger = logging.getLogger(__name__)
logger.addHandler(NullHandler())

requests_log = logging.getLogger('requests')
requests_log.setLevel(logging.WARNING)

from ably.rest.rest import AblyRest
from ably.rest.auth import Auth
from ably.types.capability import Capability
from ably.types.channeloptions import ChannelOptions
from ably.types.options import Options
from ably.util.crypto import CipherParams
from ably.util.exceptions import AblyException, AblyAuthException, IncompatibleClientIdException



from __future__ import absolute_import

import functools
import logging
import six

from ably.util.unicodemixin import UnicodeMixin

log = logging.getLogger(__name__)


class AblyException(Exception, UnicodeMixin):
    def __new__(cls, message, status_code, code):
        if cls == AblyException and status_code == 401:
            return AblyAuthException(message, status_code, code)
        return super(AblyException, cls).__new__(cls, message, status_code, code)

    def __init__(self, message, status_code, code):
        super(AblyException, self).__init__()
        self.message = message
        self.code = code
        self.status_code = status_code

    def __unicode__(self):
        return six.u('%s %s %s') % (self.code, self.status_code, self.message)

    def __str__(self):
        return self.__unicode__()

    @property
    def is_server_error(self):
        return 500 <= self.status_code <= 599

    @staticmethod
    def raise_for_response(response):
        if response.status_code >= 200 and response.status_code < 300:
            # Valid response
            return

        try:
            json_response = response.json()
        except Exception:
            log.debug("Response not json: %d %s",
                      response.status_code,
                      response.text)
            raise AblyException(message=response.text,
                                status_code=response.status_code,
                                code=response.status_code * 100)
        else:
            if json_response and 'error' in json_response:
                try:
                    raise AblyException(message=json_response['error']['message'],
                                        status_code=json_response['error']['statusCode'],
                                        code=int(json_response['error']['code']))
                except KeyError:
                    msg = "Unexpected exception decoding server response: %s"
                    msg = msg % response.text
                    raise AblyException(message=msg,
                                        status_code=500,
                                        code=50000)

            raise AblyException(message="",
                                status_code=response.status_code,
                                code=response.status_code * 100)

    @staticmethod
    def from_exception(e):
        if isinstance(e, AblyException):
            return e
        return AblyException("Unexpected exception: %s" % e, 500, 50000)


def catch_all(func):
    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        try:
            return func(*args, **kwargs)
        except Exception as e:
            log.exception(e)
            raise AblyException.from_exception(e)

    return wrapper


class AblyAuthException(AblyException):
    pass


class IncompatibleClientIdException(AblyException):
    pass



from __future__ import absolute_import

import json
import time

import six

from ably.types.capability import Capability


class TokenDetails(object):

    DEFAULTS = {'ttl': 60 * 60 * 1000}
    # Buffer in milliseconds before a token is considered unusable
    # For example, if buffer is 10000ms, the token can no longer be used for
    # new requests 9000ms before it expires
    TOKEN_EXPIRY_BUFFER = 15 * 1000

    def __init__(self, token=None, expires=None, issued=0,
                 capability=None, client_id=None):
        if expires is None:
            self.__expires = time.time() * 1000 + TokenDetails.DEFAULTS['ttl']
        else:
            self.__expires = expires
        self.__token = token
        self.__issued = issued
        if capability and isinstance(capability, six.string_types):
            self.__capability = Capability(json.loads(capability))
        else:
            self.__capability = Capability(capability or {})
        self.__client_id = client_id

    @property
    def token(self):
        return self.__token

    @property
    def expires(self):
        return self.__expires

    @property
    def issued(self):
        return self.__issued

    @property
    def capability(self):
        return self.__capability

    @property
    def client_id(self):
        return self.__client_id

    def is_expired(self, timestamp):
        if self.__expires is None:
            return False
        else:
            return self.__expires < timestamp + self.TOKEN_EXPIRY_BUFFER

    @staticmethod
    def from_dict(obj):
        kwargs = {
            'token': obj.get("token"),
            'capability': obj.get("capability"),
            'client_id': obj.get("clientId")
        }
        expires = obj.get("expires")
        kwargs['expires'] = expires if expires is None else int(expires)
        issued = obj.get("issued")
        kwargs['issued'] = issued if issued is None else int(issued)

        return TokenDetails(**kwargs)



class ChannelOptions(object):
    def __init__(self, encrypted=False, cipher_params=None):
        self.__encrypted = encrypted
        self.__cipher_params = cipher_params
        if encrypted and cipher_params is None:
            raise ValueError("Must set cipher_params if encrypted is True")

    @property
    def encrypted(self):
        return self.__encrypted

    @property
    def cipher_params(self):
        return self.__cipher_params



from __future__ import absolute_import

import logging
from datetime import datetime

import msgpack

log = logging.getLogger(__name__)


class ResourceCount(object):
    def __init__(self, opened=0, peak=0, mean=0, min=0, refused=0):
        self.opened = opened
        self.peak = peak
        self.mean = mean
        self.min = min
        self.refused = refused

    @staticmethod
    def from_dict(rc_dict):
        rc_dict = rc_dict or {}
        expected = ['opened', 'peak', 'mean', 'min', 'refused']
        kwargs = {k: rc_dict[k] for k in rc_dict if (k in expected)}

        return ResourceCount(**kwargs)


class ConnectionTypes(object):
    def __init__(self, all=None, plain=None, tls=None):
        self.all = all or ResourceCount()
        self.plain = plain or ResourceCount()
        self.tls = tls or ResourceCount()

    @staticmethod
    def from_dict(ct_dict):
        ct_dict = ct_dict or {}
        kwargs = {
            "all": ResourceCount.from_dict(ct_dict.get("all")),
            "plain": ResourceCount.from_dict(ct_dict.get("plain")),
            "tls": ResourceCount.from_dict(ct_dict.get("tls")),
        }
        return ConnectionTypes(**kwargs)


class MessageCount(object):
    def __init__(self, count=0, data=0):
        self.count = count
        self.data = data

    @staticmethod
    def from_dict(mc_dict):
        mc_dict = mc_dict or {}
        expected = ['count', 'data']
        kwargs = {k: mc_dict[k] for k in mc_dict if (k in expected)}
        return MessageCount(**kwargs)


class MessageTypes(object):
    def __init__(self, all=None, messages=None, presence=None):
        self.all = all or MessageCount()
        self.messages = messages or MessageCount()
        self.presence = presence or MessageCount()

    @staticmethod
    def from_dict(mt_dict):
        mt_dict = mt_dict or {}
        kwargs = {
            "all": MessageCount.from_dict(mt_dict.get("all")),
            "messages": MessageCount.from_dict(mt_dict.get("messages")),
            "presence": MessageCount.from_dict(mt_dict.get("presence")),
        }
        return MessageTypes(**kwargs)


class MessageTraffic(object):
    def __init__(self, all=None, realtime=None, rest=None, webhook=None):
        self.all = all or MessageTypes()
        self.realtime = realtime or MessageTypes()
        self.rest = rest or MessageTypes()
        self.webhook = webhook or MessageTypes()

    @staticmethod
    def from_dict(mt_dict):
        mt_dict = mt_dict or {}
        kwargs = {
            "all": MessageTypes.from_dict(mt_dict.get("all")),
            "realtime": MessageTypes.from_dict(mt_dict.get("realtime")),
            "rest": MessageTypes.from_dict(mt_dict.get("rest")),
            "webhook": MessageTypes.from_dict(mt_dict.get("webhook")),
        }
        return MessageTraffic(**kwargs)


class RequestCount(object):
    def __init__(self, succeeded=0, failed=0, refused=0):
        self.succeeded = succeeded
        self.failed = failed
        self.refused = refused

    @staticmethod
    def from_dict(rc_dict):
        rc_dict = rc_dict or {}
        expected = ['succeeded', 'failed', 'refused']
        kwargs = {k: rc_dict[k] for k in rc_dict if (k in expected)}
        return RequestCount(**kwargs)


class Stats(object):

    def __init__(self, all=None, inbound=None, outbound=None, persisted=None,
                 connections=None, channels=None, api_requests=None,
                 token_requests=None, interval_granularity=None,
                 interval_id=None):
        self.all = all or MessageTypes()
        self.inbound = inbound or MessageTraffic()
        self.outbound = outbound or MessageTraffic()
        self.persisted = persisted or MessageTypes()
        self.connections = connections or ConnectionTypes()
        self.channels = channels or ResourceCount()
        self.api_requests = api_requests or RequestCount()
        self.token_requests = token_requests or RequestCount()
        self.interval_id = interval_id or ''
        self.interval_granularity = (interval_granularity or
                                     granularity_from_interval_id(self.interval_id))
        self.interval_time = interval_from_interval_id(self.interval_id)

    @staticmethod
    def from_dict(stats_dict):
        stats_dict = stats_dict or {}

        kwargs = {
            "all": MessageTypes.from_dict(stats_dict.get("all")),
            "inbound": MessageTraffic.from_dict(stats_dict.get("inbound")),
            "outbound": MessageTraffic.from_dict(stats_dict.get("outbound")),
            "persisted": MessageTypes.from_dict(stats_dict.get("persisted")),
            "connections": ConnectionTypes.from_dict(stats_dict.get("connections")),
            "channels": ResourceCount.from_dict(stats_dict.get("channels")),
            "api_requests": RequestCount.from_dict(stats_dict.get("apiRequests")),
            "token_requests": RequestCount.from_dict(stats_dict.get("tokenRequests")),
            "interval_granularity": stats_dict.get("unit"),
            "interval_id": stats_dict.get("intervalId")
        }

        return Stats(**kwargs)

    @staticmethod
    def from_array(stats_array):
        return [Stats.from_dict(d) for d in stats_array]

    @staticmethod
    def to_interval_id(date_time, granularity):
        return date_time.strftime(INTERVALS_FMT[granularity])


def make_stats_response_processor(binary):
    def stats_response_processor(response):
        stats_array = response.to_native()
        return Stats.from_array(stats_array)
    return stats_response_processor


INTERVALS_FMT = {
    'minute': '%Y-%m-%d:%H:%M',
    'hour': '%Y-%m-%d:%H',
    'day': '%Y-%m-%d',
    'month': '%Y-%m',
}


def granularity_from_interval_id(interval_id):
    for key, value in INTERVALS_FMT.items():
        try:
            datetime.strptime(interval_id, value)
            return key
        except ValueError:
            pass
    raise ValueError("Unsuported intervalId")


def interval_from_interval_id(interval_id):
    granularity = granularity_from_interval_id(interval_id)
    return datetime.strptime(interval_id, INTERVALS_FMT[granularity])



# This functionality is depreceated and will be removed
# Message Pack is the replacement for all binary data messages

from __future__ import absolute_import

import json
import struct

import six


class DataType(object):
    NONE = 0
    TRUE = 1
    FALSE = 2
    INT32 = 3
    INT64 = 4
    DOUBLE = 5
    STRING = 6
    BUFFER = 7
    JSONARRAY = 8
    JSONOBJECT = 9


class Limits(object):
    INT32_MAX = 2 ** 31
    INT32_MIN = -(2 ** 31 + 1)
    INT64_MAX = 2 ** 63
    INT64_MIN = - (2 ** 63 + 1)


_decoders = {}
_decoders[DataType.TRUE] = lambda b: True
_decoders[DataType.FALSE] = lambda b: False
_decoders[DataType.INT32] = lambda b: struct.unpack('>i', b)[0]
_decoders[DataType.INT64] = lambda b: struct.unpack('>q', b)[0]
_decoders[DataType.DOUBLE] = lambda b: struct.unpack('>d', b)[0]
_decoders[DataType.STRING] = lambda b: b.decode('utf-8')
_decoders[DataType.BUFFER] = lambda b: b
_decoders[DataType.JSONARRAY] = lambda b: json.loads(b.decode('utf-8'))
_decoders[DataType.JSONOBJECT] = lambda b: json.loads(b.decode('utf-8'))


class TypedBuffer(object):
    def __init__(self, buffer, type):
        self.__buffer = buffer
        self.__type = type

    def __eq__(self, other):
        if isinstance(other, TypedBuffer):
            return self.buffer == other.buffer and self.type == other.type
        return NotImplemented

    def __ne__(self, other):
        if isinstance(other, TypedBuffer):
            result = self.__eq__(other)
            if result != NotImplemented:
                return not result
        return NotImplemented

    @staticmethod
    def from_obj(obj):
        type = DataType.NONE
        buffer = None

        if isinstance(obj, TypedBuffer):
            return obj
        elif isinstance(obj, (six.binary_type, bytearray)):
            type = DataType.BUFFER
            buffer = obj
        elif isinstance(obj, six.string_types):
            type = DataType.STRING
            buffer = obj.encode('utf-8')
        elif isinstance(obj, bool):
            type = DataType.TRUE if obj else DataType.FALSE
            buffer = None
        elif isinstance(obj, six.integer_types):
            if obj >= Limits.INT32_MIN and obj <= Limits.INT32_MAX:
                type = DataType.INT32
                buffer = struct.pack('>i', obj)
            elif obj >= Limits.INT64_MIN and obj <= Limits.INT64_MAX:
                type = DataType.INT64
                buffer = struct.pack('>q', obj)
            else:
                # TODO throw more appropriate exception
                raise 'number-too-large'
        elif isinstance(obj, float):
            type = DataType.DOUBLE
            buffer = struct.pack('>d', obj)
        elif isinstance(obj, list):
            type = DataType.JSONARRAY
            buffer = json.dumps(obj, separators=(',', ':')).encode('utf-8')
        elif isinstance(obj, dict):
            type = DataType.JSONOBJECT
            buffer = json.dumps(obj, separators=(',', ':')).encode('utf-8')
        else:
            raise 'unsupported-type'

        return TypedBuffer(buffer, type)

    @property
    def buffer(self):
        return self.__buffer

    @property
    def type(self):
        return self.__type

    def decode(self):
        decoder = _decoders[self.type]
        if decoder:
            return decoder(self.buffer)
        raise 'unsupported-type'



from __future__ import absolute_import


class HttpUtils(object):
    default_format = "json"

    mime_types = {
        "json": "application/json",
        "xml": "application/xml",
        "html": "text/html",
        "binary": "application/x-msgpack",
    }

    @staticmethod
    def default_get_headers(binary=False):
        if binary:
            return {
                "Accept": HttpUtils.mime_types['binary']
            }
        else:
            return {
                "Accept": HttpUtils.mime_types['json']
            }

    @staticmethod
    def default_post_headers(binary=False):
        if binary:
            return {
                "Accept": HttpUtils.mime_types['binary'],
                "Content-Type": HttpUtils.mime_types['binary']
            }
        else:
            return {
                "Accept": HttpUtils.mime_types['json'],
                "Content-Type": HttpUtils.mime_types['json']
            }



from __future__ import absolute_import
import random


class Defaults(object):
    protocol_version = 1
    fallback_hosts = [
        "A.ably-realtime.com",
        "B.ably-realtime.com",
        "C.ably-realtime.com",
        "D.ably-realtime.com",
        "E.ably-realtime.com",
    ]

    rest_host = "rest.ably.io"
    realtime_host = "realtime.ably.io"

    port = 80
    tls_port = 443
    connect_timeout = 15000
    disconnect_timeout = 10000
    suspended_timeout = 60000
    comet_recv_timeout = 90000
    comet_send_timeout = 10000

    transports = []  # ["web_socket", "comet"]

    @staticmethod
    def get_rest_host(options):
        if options.rest_host:
            return options.rest_host
        else:
            return Defaults.rest_host

    @staticmethod
    def get_port(options):
        if options.tls:
            if options.tls_port:
                return options.tls_port
            else:
                return Defaults.tls_port
        else:
            if options.port:
                return options.port
            else:
                return Defaults.port

    @staticmethod
    def get_fallback_rest_hosts(options):
        if options.rest_host:
            return []
        else:
            fallback_hosts_copy = list(Defaults.fallback_hosts)
            random.shuffle(fallback_hosts_copy)
            return fallback_hosts_copy

    @staticmethod
    def get_scheme(options):
        if options.tls:
            return "https"
        else:
            return "http"



from __future__ import absolute_import

import calendar
import logging
import json
from collections import OrderedDict

import six
import msgpack
from six.moves.urllib.parse import urlencode, quote

from ably.http.paginatedresult import PaginatedResult
from ably.types.message import (
    Message, make_message_response_handler, make_encrypted_message_response_handler,
    MessageJSONEncoder)
from ably.types.presence import Presence
from ably.util.crypto import get_cipher
from ably.util.exceptions import catch_all, IncompatibleClientIdException

log = logging.getLogger(__name__)


class Channel(object):
    def __init__(self, ably, name, options):
        self.__ably = ably
        self.__name = name
        self.__base_path = '/channels/%s/' % quote(name)
        self.options = options
        self.__presence = Presence(self)

    def _format_time_param(self, t):
        try:
            return '%d' % (calendar.timegm(t.utctimetuple()) * 1000)
        except:
            return '%s' % t

    @catch_all
    def history(self, direction=None, limit=None, start=None, end=None, timeout=None):
        """Returns the history for this channel"""
        params = {}

        if direction:
            params['direction'] = '%s' % direction
        if limit:
            if limit > 1000:
                raise ValueError("The maximum allowed limit is 1000")
            params['limit'] = '%d' % limit
        if start:
            params['start'] = self._format_time_param(start)
        if end:
            params['end'] = self._format_time_param(end)

        path = '/channels/%s/history' % self.__name

        if params:
            path = path + '?' + urlencode(params)

        if self.__cipher:
            message_handler = make_encrypted_message_response_handler(
                self.__cipher, self.ably.options.use_binary_protocol)
        else:
            message_handler = make_message_response_handler(
                self.ably.options.use_binary_protocol)

        return PaginatedResult.paginated_query(
            self.ably.http,
            path,
            None,
            message_handler
        )

    @catch_all
    def publish(self, name=None, data=None, client_id=None,
                messages=None, timeout=None):
        """Publishes a message on this channel.

        :Parameters:
        - `name`: the name for this message.
        - `data`: the data for this message.
        - `messages`: list of `Message` objects to be published.
            Specify this param OR `name` and `data`.

        :attention: You can publish using `name` and `data` OR `messages`, never all three.
        """
        if not messages:
            messages = [Message(name, data, client_id)]

        request_body_list = []
        for m in messages:
            if m.client_id == '*':
                raise IncompatibleClientIdException(
                    'Wildcard client_id is reserved and cannot be used when publishing messages',
                    400, 40012)
            elif m.client_id is not None and not self.ably.auth.can_assume_client_id(m.client_id):
                raise IncompatibleClientIdException(
                    'Cannot publish with client_id \'{}\' as it is incompatible with the '
                    'current configured client_id \'{}\''.format(m.client_id, self.ably.auth.client_id),
                    400, 40012)

            if self.encrypted:
                m.encrypt(self.__cipher)

            request_body_list.append(m)

        if not self.ably.options.use_binary_protocol:
            if len(request_body_list) == 1:
                request_body = request_body_list[0].as_json()
            else:
                request_body = json.dumps(request_body_list, cls=MessageJSONEncoder)
        else:
            if len(request_body_list) == 1:
                request_body = request_body_list[0].as_msgpack()
            else:
                request_body = msgpack.packb(
                    [message.as_dict(binary=True) for message in request_body_list],
                    use_bin_type=True)

        path = '/channels/%s/publish' % self.__name

        return self.ably.http.post(
            path,
            body=request_body,
            timeout=timeout
            )

    @property
    def ably(self):
        return self.__ably

    @property
    def name(self):
        return self.__name

    @property
    def base_path(self):
        return self.__base_path

    @property
    def cipher(self):
        return self.__cipher

    @property
    def encrypted(self):
        return self.options and self.options.encrypted

    @property
    def options(self):
        return self.__options

    @property
    def presence(self):
        return self.__presence

    @options.setter
    def options(self, options):
        self.__options = options

        if options and options.encrypted:
            self.__cipher = get_cipher(options.cipher_params)
        else:
            self.__cipher = None


class Channels(object):
    def __init__(self, rest):
        self.__ably = rest
        self.__attached = OrderedDict()

    def get(self, name, options=None):
        if isinstance(name, six.binary_type):
            name = name.decode('ascii')

        if name not in self.__attached:
            result = self.__attached[name] = Channel(self.__ably, name, options)
        else:
            result = self.__attached[name]
            if options is not None:
                result.options = options

        return result

    def __getitem__(self, key):
        return self.get(key)

    def __getattr__(self, name):
        try:
            return getattr(super(Channels, self), name)
        except AttributeError:
            return self.get(name)

    def __contains__(self, item):
        if isinstance(item, Channel):
            name = item.name
        elif isinstance(item, six.binary_type):
            name = item.decode('ascii')
        else:
            name = item

        return name in self.__attached

    def __iter__(self):
        return iter(six.itervalues(self.__attached))

    def release(self, key):
        del self.__attached[key]

    def __delitem__(self, key):
        return self.release(key)



from setuptools import setup

with open('LONG_DESCRIPTION.rst') as f:
    long_description = f.read()

setup(
    name='ably',
    version='0.0.2',
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: Apache Software License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.2',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Topic :: Software Development :: Libraries :: Python Modules',
    ],
    packages=['ably', 'ably.http', 'ably.rest', 'ably.transport',
              'ably.types', 'ably.util'],
    install_requires=['msgpack-python>=0.4.6',
                      'pycrypto>=2.6.1',
                      'requests>=2.7.0,<2.8',
                      'six>=1.9.0'],  # remember to update these
                                      # according to requirements.txt!
                                      # there's no easy way to reuse this.
    author="Ably",
    author_email='support@ably.io',
    url='https://github.com/ably/ably-python',
    description="A Python client library for ably.io realtime messaging",
    long_description=long_description,
)



from __future__ import absolute_import

import logging

from ably.http.http import Request
from ably.http.httputils import HttpUtils

log = logging.getLogger(__name__)


class PaginatedResult(object):
    def __init__(self, http, items, content_type, rel_first, rel_next,
                 response_processor):
        self.__http = http
        self.__items = items
        self.__content_type = content_type
        self.__rel_first = rel_first
        self.__rel_next = rel_next
        self.__response_processor = response_processor

    @property
    def items(self):
        return self.__items

    def has_first(self):
        return self.__rel_first is not None

    def has_next(self):
        return self.__rel_next is not None

    def is_last(self):
        return not self.has_next()

    def first(self):
        return self.__get_rel(self.__rel_first) if self.__rel_first else None

    def next(self):
        return self.__get_rel(self.__rel_next) if self.__rel_next else None

    def __get_rel(self, rel_req):
        if rel_req is None:
            return None
        return PaginatedResult.paginated_query_with_request(self.__http, rel_req, self.__response_processor)

    @staticmethod
    def paginated_query(http, url, headers, response_processor):
        headers = headers or {}
        req = Request(method='GET', url=url, headers=headers, body=None, skip_auth=False)
        return PaginatedResult.paginated_query_with_request(http, req, response_processor)

    @staticmethod
    def paginated_query_with_request(http, request, response_processor):
        response = http.request(request)

        items = response_processor(response)

        content_type = response.headers['Content-Type']
        links = response.links
        log.debug("Links: %s" % links)
        log.debug("Response: %s" % response)
        if 'first' in links:
            first_rel_request = request.with_relative_url(links['first']['url'])
        else:
            first_rel_request = None

        if 'next' in links:
            log.debug("Next: %s" % links['next']['url'])
            next_rel_request = request.with_relative_url(links['next']['url'])
            log.debug("Next rel request: %s" % next_rel_request)
        else:
            next_rel_request = None

        return PaginatedResult(http, items, content_type, first_rel_request,
                               next_rel_request, response_processor)



from __future__ import absolute_import

import calendar
import logging

from six.moves.urllib.parse import urlencode

from ably.http.http import Http
from ably.http.paginatedresult import PaginatedResult
from ably.rest.auth import Auth
from ably.rest.channel import Channels
from ably.util.exceptions import AblyException, catch_all
from ably.types.options import Options
from ably.types.stats import make_stats_response_processor
from ably.types.tokendetails import TokenDetails

log = logging.getLogger(__name__)


class AblyRest(object):
    """Ably Rest Client"""
    def __init__(self, key=None, token=None, token_details=None, **kwargs):
        """Create an AblyRest instance.

        :Parameters:
          **Credentials**
          - `key`: a valid key string

          **Or**
          - `token`: a valid token string
          - `token_details`: an instance of TokenDetails class

          **Optional Parameters**
          - `client_id`: Undocumented
          - `rest_host`: The host to connect to. Defaults to rest.ably.io
          - `port`: The port to connect to. Defaults to 80
          - `tls_port`: The tls_port to connect to. Defaults to 443
          - `tls`: Specifies whether the client should use TLS. Defaults
            to True
          - `auth_token`: Undocumented
          - `auth_callback`: Undocumented
          - `auth_url`: Undocumented
          - `keep_alive`: use persistent connections. Defaults to True
        """
        if key is not None and ('key_name' in kwargs or 'key_secret' in kwargs):
            raise ValueError("key and key_name or key_secret are mutually exclusive. "
                             "Provider either a key or key_name & key_secret")
        if key is not None:
            options = Options(key=key, **kwargs)
        elif token is not None:
            options = Options(auth_token=token, **kwargs)
        elif token_details is not None:
            if not isinstance(token_details, TokenDetails):
                raise ValueError("token_details must be an instance of TokenDetails")
            options = Options(token_details=token_details, **kwargs)
        elif not ('auth_callback' in kwargs or 'auth_url' in kwargs or
                  # and don't have both key_name and key_secret
                  ('key_name' in kwargs and 'key_secret' in kwargs)):
            raise ValueError("key is missing. Either an API key, token, or token auth method must be provided")
        else:
            options = Options(**kwargs)

        # if self.__keep_alive:
        #     self.__session = requests.Session()
        # else:
        #     self.__session = None

        self.__http = Http(self, options)
        self.__auth = Auth(self, options)
        self.__http.auth = self.__auth

        self.__channels = Channels(self)
        self.__options = options

    def _format_time_param(self, t):
        try:
            return '%d' % (calendar.timegm(t.utctimetuple()) * 1000)
        except:
            return '%s' % t

    @catch_all
    def stats(self, direction=None, start=None, end=None, params=None,
              limit=None, paginated=None, unit=None, timeout=None):
        """Returns the stats for this application"""
        params = params or {}

        if direction:
            params["direction"] = direction
        if start:
            params["start"] = self._format_time_param(start)
        if end:
            params["end"] = self._format_time_param(end)
        if limit:
            if limit > 1000:
                raise ValueError("The maximum allowed limit is 1000")
            params["limit"] = limit
        if unit:
            params["unit"] = unit

        if 'start' in params and 'end' in params and params['start'] > params['end']:
            raise ValueError("'end' parameter has to be greater than or equal to 'start'")

        url = '/stats'
        if params:
            url += '?' + urlencode(params)

        stats_response_processor = make_stats_response_processor(
            self.options.use_binary_protocol)

        return PaginatedResult.paginated_query(self.http,
                                               url, None,
                                               stats_response_processor)

    @catch_all
    def time(self, timeout=None):
        """Returns the current server time in ms since the unix epoch"""
        r = self.http.get('/time', skip_auth=True, timeout=timeout)
        AblyException.raise_for_response(r)
        return r.to_native()[0]

    @property
    def client_id(self):
        return self.options.client_id

    @property
    def channels(self):
        """Returns the channels container object"""
        return self.__channels

    @property
    def auth(self):
        return self.__auth

    @property
    def http(self):
        return self.__http

    @property
    def options(self):
        return self.__options



from __future__ import absolute_import

import functools
import itertools
import logging
import time
import json

from six.moves import range
from six.moves.urllib.parse import urljoin

import requests
import msgpack

from ably.rest.auth import Auth
from ably.http.httputils import HttpUtils
from ably.transport.defaults import Defaults
from ably.util.exceptions import AblyException, AblyAuthException

log = logging.getLogger(__name__)


def reauth_if_expired(func):
    @functools.wraps(func)
    def wrapper(rest, *args, **kwargs):
        if kwargs.get("skip_auth"):
            return func(rest, *args, **kwargs)

        num_tries = 5
        for i in range(num_tries):
            try:
                return func(rest, *args, **kwargs)
            except AblyException as e:
                if e.code == 40140 and i < (num_tries - 1):
                    rest.reauth()
                    continue
                raise
    return wrapper


class Request(object):
    def __init__(self, method='GET', url='/', headers=None, body=None, skip_auth=False):
        self.__method = method
        self.__headers = headers or {}
        self.__body = body
        self.__skip_auth = skip_auth
        self.__url = url

    def with_relative_url(self, relative_url):
        return Request(self.method, urljoin(self.url, relative_url), self.headers, self.body, self.skip_auth)

    @property
    def method(self):
        return self.__method

    @property
    def url(self):
        return self.__url

    @property
    def headers(self):
        return self.__headers

    @property
    def body(self):
        return self.__body

    @property
    def skip_auth(self):
        return self.__skip_auth


class Response(object):
    """
    Composition for requests.Response with delegation
    """

    def __init__(self, response):
        self.__response = response

    def to_native(self):
        content_type = self.__response.headers.get('content-type')
        if content_type == 'application/x-msgpack':
            return msgpack.unpackb(self.__response.content, encoding='utf-8')
        elif content_type == 'application/json':
            return self.json()
        else:
            raise ValueError("Unsuported content type")

    def __getattr__(self, attr):
        return getattr(self.__response, attr)


class Http(object):
    CONNECTION_RETRY_DEFAULTS = {
        'http_open_timeout': 4,
        'http_request_timeout': 15,
        'http_max_retry_count': 3,
        'http_max_retry_duration': 10,
    }

    def __init__(self, ably, options):
        options = options or {}
        self.__ably = ably
        self.__options = options

        self.__session = requests.Session()
        self.__auth = None

    def dump_body(self, body):
        if self.options.use_binary_protocol:
            return msgpack.packb(body, use_bin_type=False)
        else:
            return json.dumps(body, separators=(',', ':'))

    def reauth(self):
        try:
            self.auth.authorise(force=True)
        except AblyAuthException as e:
            if e.code == 40101:
                e.message = ("The provided token is not renewable and there is"
                             " no means to generate a new token")
            raise e

    @reauth_if_expired
    def make_request(self, method, path, headers=None, body=None,
                     native_data=None, skip_auth=False, timeout=None):
        fallback_hosts = Defaults.get_fallback_rest_hosts(self.__options)
        if fallback_hosts:
            fallback_hosts.insert(0, self.preferred_host)
            fallback_hosts = itertools.cycle(fallback_hosts)
        if native_data is not None and body is not None:
            raise ValueError("make_request takes either body or native_data")
        elif native_data is not None:
            body = self.dump_body(native_data)
        if body:
            all_headers = HttpUtils.default_post_headers(
                self.options.use_binary_protocol)
        else:
            all_headers = HttpUtils.default_get_headers(
                self.options.use_binary_protocol)

        if not skip_auth:
            if self.auth.auth_mechanism == Auth.Method.BASIC and self.preferred_scheme.lower() == 'http':
                raise AblyException(
                    "Cannot use Basic Auth over non-TLS connections",
                    401,
                    40103)
            all_headers.update(self.auth._get_auth_headers())
        if headers:
            all_headers.update(headers)

        http_open_timeout = self.http_open_timeout
        http_request_timeout = self.http_request_timeout
        if fallback_hosts:
            http_max_retry_count = self.http_max_retry_count
        else:
            http_max_retry_count = 1
        http_max_retry_duration = self.http_max_retry_duration
        requested_at = time.time()
        for retry_count in range(http_max_retry_count):
            host = next(fallback_hosts) if fallback_hosts else self.preferred_host
            if self.options.environment:
                host = self.options.environment + '-' + host

            base_url = "%s://%s:%d" % (self.preferred_scheme,
                                       host,
                                       self.preferred_port)
            url = urljoin(base_url, path)
            request = requests.Request(method, url, data=body, headers=all_headers)
            prepped = self.__session.prepare_request(request)
            try:
                response = self.__session.send(
                    prepped,
                    timeout=(http_open_timeout,
                             http_request_timeout))
            except Exception as e:
                # Need to catch `Exception`, see:
                # https://github.com/kennethreitz/requests/issues/1236#issuecomment-133312626

                # if last try or cumulative timeout is done, throw exception up
                time_passed = time.time() - requested_at
                if retry_count == http_max_retry_count - 1 or \
                   time_passed > http_max_retry_duration:
                    raise e
            else:
                try:
                    AblyException.raise_for_response(response)
                    return Response(response)
                except AblyException as e:
                    if not e.is_server_error:
                        raise e

    def request(self, request):
        return self.make_request(request.method, request.url, headers=request.headers, body=request.body,
                                 skip_auth=request.skip_auth)

    def get(self, url, headers=None, skip_auth=False, timeout=None):
        return self.make_request('GET', url, headers=headers, skip_auth=skip_auth, timeout=timeout)

    def post(self, url, headers=None, body=None, native_data=None, skip_auth=False, timeout=None):
        return self.make_request('POST', url, headers=headers, body=body, native_data=native_data,
                                 skip_auth=skip_auth, timeout=timeout)

    def delete(self, url, headers=None, skip_auth=False, timeout=None):
        return self.make_request('DELETE', url, headers=headers, skip_auth=skip_auth, timeout=timeout)

    @property
    def auth(self):
        return self.__auth

    @auth.setter
    def auth(self, value):
        self.__auth = value

    @property
    def options(self):
        return self.__options

    @property
    def preferred_host(self):
        return Defaults.get_rest_host(self.options)

    @property
    def preferred_port(self):
        return Defaults.get_port(self.options)

    @property
    def preferred_scheme(self):
        return Defaults.get_scheme(self.options)

    @property
    def http_open_timeout(self):
        if self.options.http_open_timeout is not None:
            return self.options.http_open_timeout
        return self.CONNECTION_RETRY_DEFAULTS['http_open_timeout']

    @property
    def http_request_timeout(self):
        if self.options.http_request_timeout is not None:
            return self.options.http_request_timeout
        return self.CONNECTION_RETRY_DEFAULTS['http_request_timeout']

    @property
    def http_max_retry_count(self):
        if self.options.http_max_retry_count is not None:
            return self.options.http_max_retry_count
        return self.CONNECTION_RETRY_DEFAULTS['http_max_retry_count']

    @property
    def http_max_retry_duration(self):
        if self.options.http_max_retry_duration is not None:
            return self.options.http_max_retry_duration
        return self.CONNECTION_RETRY_DEFAULTS['http_max_retry_duration']



from __future__ import absolute_import

from datetime import datetime, timedelta

from six.moves.urllib.parse import urlencode

from ably.http.paginatedresult import PaginatedResult
from ably.types.mixins import EncodeDataMixin


def _ms_since_epoch(dt):
    epoch = datetime.utcfromtimestamp(0)
    delta = dt - epoch
    return int(delta.total_seconds() * 1000)


def _dt_from_ms_epoch(ms):
    epoch = datetime.utcfromtimestamp(0)
    return epoch + timedelta(milliseconds=ms)


class PresenceAction(object):
    ABSENT = 0
    PRESENT = 1
    ENTER = 2
    LEAVE = 3
    UPDATE = 4


class PresenceMessage(EncodeDataMixin):
    def __init__(self, id=None, action=None, client_id=None,
                 data=None, encoding=None, connection_id=None,
                 timestamp=None):
        self.__id = id
        self.__action = action
        self.__client_id = client_id
        self.__connection_id = connection_id
        self.__data = data
        self.__encoding = encoding
        self.__timestamp = timestamp

    @staticmethod
    def from_dict(obj, cipher=None):
        id = obj.get('id')
        action = obj.get('action', PresenceAction.ENTER)
        client_id = obj.get('clientId')
        connection_id = obj.get('connectionId')

        encoding = obj.get('encoding', '')
        timestamp = obj.get('timestamp')

        if timestamp is not None:
            timestamp = _dt_from_ms_epoch(timestamp)

        data = obj.get('data')

        decoded_data = PresenceMessage.decode(data, encoding, cipher)

        return PresenceMessage(
            id=id,
            action=action,
            client_id=client_id,
            connection_id=connection_id,
            timestamp=timestamp,
            **decoded_data
        )

    @staticmethod
    def messages_from_array(obj, cipher=None):
        return [PresenceMessage.from_dict(d, cipher) for d in obj]

    @property
    def action(self):
        return self.__action

    @property
    def client_id(self):
        return self.__client_id

    @property
    def encoding(self):
        return self.__encoding

    @property
    def member_key(self):
        if self.connection_id and self.client_id:
            return "%s:%s" % (self.connection_id, self.client_id)

    @property
    def data(self):
        return self.__data

    @property
    def id(self):
        return self.__id

    @property
    def connection_id(self):
        return self.__connection_id

    @property
    def timestamp(self):
        return self.__timestamp


class Presence(object):
    def __init__(self, channel):
        self.__base_path = channel.base_path
        self.__binary = channel.ably.options.use_binary_protocol
        self.__http = channel.ably.http
        self.__cipher = channel.cipher

    def _path_with_qs(self, rel_path, qs=None):
        path = rel_path
        if qs:
            path += ('?' + urlencode(qs))
        return path

    def get(self, limit=None):
        qs = {}
        if limit:
            if limit > 1000:
                raise ValueError("The maximum allowed limit is 1000")
            qs['limit'] = limit
        path = self._path_with_qs('%s/presence' % self.__base_path.rstrip('/'), qs)

        if self.__cipher:
            presence_handler = make_encrypted_presence_response_handler(self.__cipher, self.__binary)
        else:
            presence_handler = make_presence_response_handler(self.__binary)

        return PaginatedResult.paginated_query(
            self.__http,
            path,
            {},
            presence_handler)

    def history(self, limit=None, direction=None, start=None, end=None):
        qs = {}
        if limit:
            if limit > 1000:
                raise ValueError("The maximum allowed limit is 1000")
            qs['limit'] = limit
        if direction:
            qs['direction'] = direction
        if start:
            if isinstance(start, int):
                qs['start'] = start
            else:
                qs['start'] = _ms_since_epoch(start)
        if end:
            if isinstance(end, int):
                qs['end'] = end
            else:
                qs['end'] = _ms_since_epoch(end)

        if 'start' in qs and 'end' in qs and qs['start'] > qs['end']:
            raise ValueError("'end' parameter has to be greater than or equal to 'start'")

        path = self._path_with_qs('%s/presence/history' % self.__base_path.rstrip('/'), qs)

        if self.__cipher:
            presence_handler = make_encrypted_presence_response_handler(
                self.__cipher, self.__binary)
        else:
            presence_handler = make_presence_response_handler(self.__binary)

        return PaginatedResult.paginated_query(
            self.__http,
            path,
            {},
            presence_handler
        )


def make_presence_response_handler(binary):
    def presence_response_handler(response):
        messages = response.to_native()
        return [PresenceMessage.from_dict(message) for message in messages]
    return presence_response_handler


def make_encrypted_presence_response_handler(cipher, binary):
    def encrypted_presence_response_handler(response):
        messages = response.to_native()
        return [PresenceMessage.from_dict(message, cipher) for message in messages]
    return encrypted_presence_response_handler




import base64

import six

import hashlib
import hmac


class TokenRequest(object):

    def __init__(self, key_name=None, client_id=None, nonce=None, mac=None,
                 capability=None, ttl=None, timestamp=None):
        self.__key_name = key_name
        self.__client_id = client_id
        self.__nonce = nonce
        self.__mac = mac
        self.__capability = capability
        self.__ttl = ttl
        self.__timestamp = timestamp

    def sign_request(self, key_secret):
        sign_text = six.u("\n").join([six.text_type(x) for x in [
            self.key_name or "",
            self.ttl or "",
            self.capability or "",
            self.client_id or "",
            "%d" % (self.timestamp or 0),
            self.nonce or "",
            "",  # to get the trailing new line
        ]])
        try:
            key_secret = key_secret.encode('utf8')
        except AttributeError:
            pass
        try:
            sign_text = sign_text.encode('utf8')
        except AttributeError:
            pass
        mac = hmac.new(key_secret, sign_text, hashlib.sha256).digest()
        self.mac = base64.b64encode(mac).decode('utf8')

    def to_dict(self):
        return {
            'keyName': self.key_name,
            'clientId': self.client_id,
            'ttl': self.ttl,
            'nonce': self.nonce,
            'capability': self.capability,
            'timestamp': self.timestamp,
            'mac': self.mac
        }

    @property
    def key_name(self):
        return self.__key_name

    @property
    def client_id(self):
        return self.__client_id

    @property
    def nonce(self):
        return self.__nonce

    @property
    def mac(self):
        return self.__mac

    @mac.setter
    def mac(self, mac):
        self.__mac = mac

    @property
    def capability(self):
        return self.__capability

    @property
    def ttl(self):
        return self.__ttl

    @property
    def timestamp(self):
        return self.__timestamp
